#!/usr/bin/env python
# Modified from Chu, S. X. (2021). Data for "Fault interactions enhance high-frequency earthquake radiation" [Data set]. Zenodo. https://doi.org/10.5281/zenodo.5104902

from shapely import affinity
from shapely.ops import unary_union
import geopandas as gpd
import numpy as np

def calc_rd(faults: gpd.GeoDataFrame) -> float:  
    """
    Calculate the density ratio (Rd) of summed fault lengths to their convex hull's perimeter.
    
    Parameters:
    faults (gpd.GeoDataFrame): GeoDataFrame containing fault line geometries.
    
    Returns:
    float: The ratio of the total length of all fault lines to the perimeter of their convex hull.
           Returns 0.0 if the convex hull has a length of zero.
    """
    faults = unary_union(faults.geometry)
    hull_length = faults.convex_hull.length
    total_length = faults.length
    return total_length / hull_length if hull_length != 0 else 0.0

def projected_length(faults: gpd.GeoDataFrame, angle: float) -> float:
    """
    Calculate the summed projection length of individual faults to a given angle.
    
    Parameters:
    faults (gpd.GeoDataFrame): GeoDataFrame containing fault line geometries.
    angle (float): The angle (in degrees) to project the fault lines.
    
    Returns:
    float: The summed projection length of fault lines at the given angle.
    """
    projection = 0.0
    for segment in faults.geometry:
        rotated_segment = affinity.rotate(segment, -angle)
        _, minY, _, maxY = rotated_segment.bounds
        projection += abs(maxY - minY)
    return projection

def calc_rm(faults: gpd.GeoDataFrame) -> (float, float, float):
    """
    Calculate the misalignment ratio (Rm) and the angles with the maximum and minimum summed projection length.
    
    Parameters:
    faults (gpd.GeoDataFrame): GeoDataFrame containing fault line geometries.
    
    Returns:
    tuple: A tuple containing the misalignment ratio and the angles of maximum and minimum summed projection length.
           Returns (0.0, 0.0, 0.0) if the maximum projection is zero.
    """
    degrees = np.arange(0., 180., 1)
    projections = [projected_length(faults, angle) for angle in degrees]
    max_proj = max(projections)
    min_proj = min(projections)
    if max_proj == 0:
        return 0.0, 0.0, 0.0
    max_angle = degrees[np.argmax(projections)]
    min_angle = degrees[np.argmin(projections)]
    return (min_proj / max_proj, max_angle, min_angle)

def calculate_RmRd(zone: gpd.GeoDataFrame, fault_map: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """
    Calculate fault misalignment (Rm) and fault density ratios (Rd) for each zone using fault data.
    
    Parameters:
    zone (gpd.GeoDataFrame): GeoDataFrame containing zones to calculate the ratios for.
    fault_map (gpd.GeoDataFrame): GeoDataFrame containing fault lines used for calculations.
    
    Returns:
    gpd.GeoDataFrame: The input GeoDataFrame with new columns 'Rm' and 'Rd' containing the calculated ratios.
    """
    zone['Rm'] = 0.0 
    zone['Rd'] = 0.0
    for index, area in zone.iterrows():
        faults = gpd.clip(fault_map, area.geometry)
        if not faults.empty:
            Rm, max_angle, min_angle = calc_rm(faults)
            zone.at[index, 'Rm'] = Rm
            zone.at[index, 'Rd'] = calc_rd(faults)
    return zone